@@ -1,6 +1,5 @@
 #!/usr/bin/env python
 
-from appdirs import user_cache_dir
 import argparse
 import fnmatch
 import json
@@ -12,26 +11,33 @@ import shutil
 import signal
 import subprocess
 import sys
+import re
 import tarfile
 import tempfile
 import textwrap
 import time
 from urllib.request import urlopen
 
-__version__ = "0.99"
+__version__ = "1.0"
 
 DOCKER_NS = "satex"
 REGISTRY_URL = "https://github.com/sat-heritage/docker-images/releases/download/list/list.tgz"
 
-cache_validity = 3600*4
-cache_dir = user_cache_dir("satex", "satex")
-cache_file = os.path.join(cache_dir, "list.tgz")
-
 on_linux = platform.system() == "Linux"
 
-def error(msg):
-    print(msg, file=sys.stderr)
-    sys.exit(1)
+def color(s, color, mode=1):
+    return f"\033[{mode};{color}m{s}\033[0m"
+def red(s):
+    return color(s, 31)
+def green(s):
+    return color(s, 32)
+def yellow(s):
+    return color(s, 33)
+
+def error(msg, exit=True):
+    print("\033[1;31mERROR %s\033[0m" % msg, file=sys.stderr)
+    if exit:
+        sys.exit(1)
 
 def warn(msg):
     print("\033[1;33m! %s\033[0m" % msg, file=sys.stderr)
@@ -60,6 +66,12 @@ def fetch_registry(args, opener):
             cfg[tag] = json.load(fp)
     return reg, cfg
 
+if not IN_REPOSITORY:
+    from appdirs import user_cache_dir
+    cache_validity = 3600*4
+    cache_dir = user_cache_dir("satex", "satex")
+    cache_file = os.path.join(cache_dir, "list.tgz")
+
 def is_cache_valid(args):
     if args.refresh_list:
         return False
@@ -94,6 +106,13 @@ def make_name(reg, cfg, entry, solver):
     pattern = cfg[entry].get("image_name", "{SOLVER}:{ENTRY}")
     return pattern.format(ENTRY=entry, SOLVER=solver)
 
+def is_no_pattern(spec):
+    return not set(spec).intersection("?[*")
+
+re_image_name = re.compile("[a-zA-Z0-9\-_\.]+:[a-zA-Z0-9\-_\.]+")
+def valid_name(name):
+    return re_image_name.match(name)
+
 class Repository(object):
     def __init__(self, args):
         self.registry, self.setup = get_registry(args)
@@ -101,13 +120,19 @@ class Repository(object):
         self.names = []
 
         select_all = not hasattr(args, "all") or args.all
+        if hasattr(args, "pattern") and is_no_pattern(args.pattern):
+            select_all = True
         select_fixme = select_all or hasattr(args, "fixme") and args.fixme
         select_unstable = select_all or hasattr(args, "unstable") and args.unstable
         select_stable = select_all or (not select_unstable and not select_fixme)
 
+        select_tracks = set([args.track]) if hasattr(args, "track") and args.track else None
+
         for entry in self.registry:
             for solver in self.registry[entry]:
                 name = make_name(self.registry, self.setup, entry, solver)
+                if not valid_name(name):
+                    error(f"invalid image name: '{name}'")
                 if hasattr(args, "pattern") and \
                         not fnmatch.fnmatch(name, args.pattern):
                     continue
@@ -120,6 +145,11 @@ class Repository(object):
                         continue
                 elif not select_unstable:
                     continue
+
+                tracks = self.registry[entry][solver].get("tracks", [])
+                if select_tracks and not select_tracks.intersection(tracks):
+                    continue
+
                 self.images[name] = {"entry": entry, "solver": solver}
                 self.names.append(name)
 
@@ -196,7 +226,10 @@ def print_info(args):
                     name += " (proof)"
                 value = f"{image.registry['call']} {' '.join(map(str,value))}"
             else:
-                value = str(value)
+                if isinstance(value, list):
+                    value = ", ".join(map(str, value))
+                else:
+                    value = str(value)
             key_width = max(len(name), key_width)
             info.append({"key": key, "name": name, "value": value})
 
@@ -254,7 +287,7 @@ def docker_call():
         except KeyError:
             raise
         if not check_sudo():
-            error("""Error: 'sudo' is not installed and you are not in the 'docker' group.
+            error("""'sudo' is not installed and you are not in the 'docker' group.
 Either install sudo, or add your user to the docker group by doing
    su -c "usermod -aG docker $USER" """)
         return sudo_docker
@@ -263,11 +296,11 @@ Either install sudo, or add your user to the docker group by doing
 def check_docker():
     if not check_cmd(["docker", "version"]):
         if not on_linux:
-            error("""Error: Docker not found.
+            error("""Docker not found.
 If you are using Docker Toolbox, make sure you are running 'satex'
 within the 'Docker quickstart Terminal'.""")
         else:
-            error("Error: Docker not found.")
+            error("Docker not found.")
     docker_argv = docker_call()
     #if not check_cmd(docker_argv + ["version"]):
     #    error("Error: cannot connect to Docker. Make sure it is running.")
@@ -280,6 +313,14 @@ def prepare_image(args, docker_argv, image):
         info(" ".join(cmd))
         subprocess.check_call(cmd)
 
+def easy_volume(v):
+    if ":" in v:
+        orig, dest = v.split(":")
+        if orig[0] != "/" and os.path.isdir(orig):
+            orig = os.path.abspath(orig)
+        v = f"{orig}:{dest}"
+    return v
+
 _docker_opts = []
 def docker_runs(args, images, docker_args=(), image_args=()):
     docker_argv = check_docker()
@@ -287,6 +328,7 @@ def docker_runs(args, images, docker_args=(), image_args=()):
     argv = ["run", "--name", container_id, "--rm"]
     if hasattr(args, "timeout"):
         argv += ["-e", f"TIMEOUT={args.timeout}"]
+    quiet = hasattr(args, "quiet") and args.quiet
     for opt in _docker_opts:
         if getattr(args, opt) is not None:
             val = getattr(args, opt)
@@ -298,7 +340,14 @@ def docker_runs(args, images, docker_args=(), image_args=()):
             else:
                 argv += ["--%s" % opt, val]
     argv += list(docker_args)
-    image_argv = list(image_args)
+    image_argv = ["--mode", args.mode] if hasattr(args, "mode") and args.mode else []
+    image_argv += list(image_args)
+    run_args = {}
+    if quiet:
+        run_args["stdout"] = subprocess.DEVNULL
+        run_args["stderr"] = subprocess.DEVNULL
+    global stop
+    stop = False
     for image in images:
         image = f"{DOCKER_NS}/{image}"
         prepare_image(args, docker_argv, image)
@@ -307,20 +356,26 @@ def docker_runs(args, images, docker_args=(), image_args=()):
             print(" ".join(cmd))
         else:
             def killer(s,f):
+                global stop
+                stop = True
                 warn("Killing solver...")
                 argv = docker_argv + ["kill", container_id]
                 subprocess.run(argv, stdout=subprocess.DEVNULL)
             signal.signal(signal.SIGINT, killer)
-            info(" ".join(cmd))
-            ret = subprocess.run(cmd).returncode
+            info(" ".join(cmd)) if not quiet else None
+            ret = subprocess.run(cmd, **run_args).returncode
             signal.signal(signal.SIGINT, signal.SIG_DFL)
+            if stop:
+                sys.exit(1)
             if ret == 124:
                 if args.fail_if_timeout:
                     raise subprocess.TimeoutExpired(image, args.timeout)
-                else:
+                elif not quiet:
                     warn(f"{image} timeout")
             else:
-                assert ret == 0 or 10 <= ret <= 20, f"Solver failed! ({ret})"
+                if not (ret == 0 or 10 <= ret <= 20):
+                    if not quiet:
+                        error(f"Solver failed with return code {ret}")
     if not args.pretend:
         return ret
 
@@ -484,13 +539,77 @@ def build_images(args):
             os.unlink(dbjson)
 
 
+_retstr = {
+    10: "SAT",
+    20: "UNSAT",
+    124: "TIMEOUT",
+}
+
 def test_images(args):
-    images = get_list(args)
     docker_args = ["-v", f"{os.path.abspath('tests')}:/data"]
-    dimacs = args.cnf
-    info(f"Testing {dimacs}")
-    dimacs = os.path.basename(dimacs)
-    docker_runs(args, images, docker_args, image_args=(dimacs,))
+
+    info(f"Testing with CNF {os.path.abspath('tests')}/{args.file}")
+
+    if args.timeout > 600:
+        args.timeout = 10
+
+    def call(test_name, image, image_args):
+        print(test_name, end="...", flush=True)
+        ret = docker_runs(args, [image.name], docker_args=docker_args,
+                    image_args=image_args)
+        msg = _retstr.get(ret, ret)
+        if ret == 124 or ret == 0 or 10 <= ret <= 20:
+            print(green("ok"), f"({msg})")
+            return True
+        else:
+            print(red("fail"), f"({msg})")
+            return False
+
+    def test_cnf(image):
+        image_args = [args.file]
+        return call("cnf", image, image_args)
+    def test_gz(image):
+        image_args = [f"{args.file}.gz"]
+        return call("gz", image, image_args)
+    def test_proof(image):
+        if not "argsproof" in image.registry:
+            return True
+        image_args = [args.file, "proof.tmp"]
+        return call("proof", image, image_args)
+    def test_modes(image):
+        ok = True
+        for mode in [k for k in image.registry if k.startswith("args")]:
+            mode = mode[4:]
+            if not mode or mode == "proof":
+                continue
+            image_args = ["--mode", mode[4:], "aim-200-1_6-yes1-1.cnf.gz", "proof.tmp"]
+            ok = call(mode, image, image_args) and ok
+        return ok
+
+    tests = [test_cnf, test_gz, test_proof, test_modes]
+
+    failures = []
+
+    repo = Repository(args)
+    docker_argv = check_docker()
+    for name in repo.images:
+        prepare_image(args, docker_argv, f"{DOCKER_NS}/{name}")
+    for name in repo.images:
+        image = ImageManager(name, repo)
+        info(f"Testing {image.name}")
+        fails = [test.__name__[5:] for test in tests if not test(image)]
+        if fails:
+            failures.append((image, fails))
+
+    if not failures:
+        print(green("Bravo :-)"))
+    else:
+        print(red("Failed tests:"))
+        for image, fails in failures:
+            print(f"{lightred(image.name)} failed tests {' '.join(fails)}")
+
+    if failures:
+        sys.exit(1)
 
 def push_images(args):
     docker_argv = check_docker()
@@ -558,11 +677,13 @@ def main(redirected=False):
     status_parser.add_argument("--all", "-a", action="store_true",
             help="Consider all images, with any status")
 
-    spec_parser = argparse.ArgumentParser(add_help=False,
-            parents=[status_parser])
+    spec_parser = argparse.ArgumentParser(add_help=False)
     spec_parser.add_argument("pattern",
             help="Pattern for filtering images")
 
+    tracks_parser = argparse.ArgumentParser(add_help=False)
+    tracks_parser.add_argument("--track",
+            help="Filter solvers from the given track")
 
     docker_parser = argparse.ArgumentParser(add_help=False)
     docker_parser.add_argument("--pull", action="store_true",
@@ -588,19 +709,22 @@ def main(redirected=False):
 
     p = subparsers.add_parser("list",
             help=f"List {DOCKER_NS} Docker images",
-            parents=[status_parser])
+            parents=[status_parser, tracks_parser])
     p.add_argument("pattern", default="*", nargs="?",
             help="Pattern for filtering images (default: *)")
     p.set_defaults(func=print_list)
 
     p = subparsers.add_parser("info",
             help=f"Display information about the solver embedded in the given Docker images",
-            parents=[spec_parser])
+            parents=[spec_parser, tracks_parser])
     p.set_defaults(func=print_info)
 
     p = subparsers.add_parser("run",
             help=f"Run one or several {DOCKER_NS} Docker images",
-            parents=[spec_parser, run_parser, docker_parser])
+            parents=[spec_parser, status_parser, tracks_parser,
+                run_parser, docker_parser])
+    p.add_argument("--mode",
+            help="Select args mode")
     p.add_argument("dimacs",
             help="DIMACS file (possibly gzipped)")
     p.add_argument("proof", nargs="?",
@@ -609,7 +733,8 @@ def main(redirected=False):
 
     p = subparsers.add_parser("run-raw",
             help=f"Run one or several {DOCKER_NS} Docker images with direct call to solvers",
-            parents=[spec_parser, run_parser, docker_parser])
+            parents=[spec_parser, status_parser, tracks_parser,
+                run_parser, docker_parser])
     p.add_argument("args", nargs=argparse.REMAINDER,
             help="Arguments to docker image")
     p.set_defaults(func=runraw_images)
@@ -622,7 +747,7 @@ def main(redirected=False):
 
     p = subparsers.add_parser("extract",
             help=f"Extract solvers binaries from {DOCKER_NS} Docker images",
-            parents=[spec_parser, docker_parser])
+            parents=[spec_parser, status_parser, tracks_parser, docker_parser])
     p.add_argument("output_dir", help="Output directory")
     p.set_defaults(func=extract)
 
@@ -635,20 +760,22 @@ def main(redirected=False):
     if IN_REPOSITORY:
         p = subparsers.add_parser("build",
                 help=f"Build {DOCKER_NS} Docker images",
-                parents=[spec_parser])
+                parents=[spec_parser, status_parser, tracks_parser])
         p.add_argument("--no-cache", action="store_true",
                 help="docker build option")
         p.set_defaults(func=build_images)
 
         p = subparsers.add_parser("test",
                 help=f"Test {DOCKER_NS} Docker images",
-                parents=[spec_parser, run_parser, docker_parser])
-        p.add_argument("--cnf", "-f", default="tests/cmu-bmc-barrel6.cnf.gz")
+                parents=[spec_parser, status_parser, run_parser, tracks_parser, docker_parser])
+        p.add_argument("--quiet", "-q", action="store_true")
+        p.add_argument("--file", "-f", default="aim-200-1_6-yes1-1.cnf",
+                help=".cnf test file (should also exists with .gz)")
         p.set_defaults(func=test_images)
 
         p = subparsers.add_parser("push",
                 help=f"Push {DOCKER_NS} Docker images",
-                parents=[spec_parser])
+                parents=[spec_parser, status_parser, tracks_parser])
         p.set_defaults(func=push_images)
 
     subparsers.add_parser("version",
